package com.mindoo.ftp.logging;

import java.io.PrintWriter;
import java.io.StringWriter;
import java.util.Date;
import java.util.concurrent.BlockingQueue;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.logging.Level;

import com.mindoo.ftp.config.nsf.FTPDbLocator;

import lotus.domino.Database;
import lotus.domino.DateTime;
import lotus.domino.Document;
import lotus.domino.Item;
import lotus.domino.Name;
import lotus.domino.NotesException;
import lotus.domino.NotesFactory;
import lotus.domino.NotesThread;
import lotus.domino.Session;

/**
 * Background thread that asynchronously writes log database entries.
 * If the log database cannot be located, the log messages are written
 * to the server console.
 * 
 * @author Karsten Lehmann
 */
public class LogNotesThread extends NotesThread {
	private BlockingQueue<LogData> m_pendingLogEntries=new LinkedBlockingQueue<LogData>();

	/**
	 * Creates a new thread instance
	 */
	public LogNotesThread() {
		super("Mindoo FTP server log thread");
	}
	
	/**
	 * Adds a new entry to the log
	 * 
	 * @param level log level
	 * @param message log message
	 * @param t optional exception or <code>null</code>
	 */
	public void log(Level level, String message, Throwable t) {
		LogData data=new LogData();
		data.m_level=level;
		data.m_message=message;
		data.m_exception=t;
		m_pendingLogEntries.add(data);
	}

	/**
	 * Method notifies the thread to shut down gracefully
	 */
	public void shutdown() {
		m_pendingLogEntries.add(new ShutdownLogData());
	}
	
	@Override
	public void runNotes() throws NotesException {
		Session session=NotesFactory.createSession();
		Database logDb=FTPDbLocator.getLogDatabase(session);
		Name serverName=session.createName(session.getServerName());

		logDb=FTPDbLocator.getLogDatabase(session);
		
		while (true) {
			try {
				LogData data=m_pendingLogEntries.take();
				
				//check if this thread should shut down
				if (data instanceof ShutdownLogData)
					break;
				
				Level level=data.m_level;
				String message=data.m_message;
				Throwable t=data.m_exception;
				String exceptionAsString=null;
				if (t!=null) {
					StringWriter sWriter=new StringWriter();
					PrintWriter pWriter=new PrintWriter(sWriter);
					t.printStackTrace(pWriter);
					exceptionAsString=sWriter.toString();
				}
				
				if (logDb==null) {
					//log to console
					System.out.println((new Date()).toString()+" "+level.getLocalizedName()+" - "+message+(exceptionAsString==null || exceptionAsString.length()==0 ? "" : "\n"+exceptionAsString));
				}
				else {
					//log to db
					Document logDoc=logDb.createDocument();
					logDoc.replaceItemValue("Form", "LogEntry");
					Item serverNameItm=logDoc.replaceItemValue("Server", serverName.getCanonical());
					serverNameItm.setNames(true);
					
					DateTime ndtNow=session.createDateTime(new Date());
					logDoc.replaceItemValue("Created", ndtNow);
					ndtNow.recycle();
					logDoc.replaceItemValue("Level", level.intValue());
					logDoc.replaceItemValue("Message", message);
					if (exceptionAsString!=null) {
						logDoc.replaceItemValue("Exception", exceptionAsString);
					}
					logDoc.save();
					logDoc.recycle();
				}
			}
			catch (InterruptedException e) {
				e.printStackTrace();
			}
		}
		session.recycle();
	}
	
	/**
	 * Container for log data
	 * 
	 * @author Karsten Lehmann
	 */
	private static class LogData {
		public Level m_level;
		public String m_message;
		public Throwable m_exception;
	}
	
	/**
	 * Shutdown request for Notes thread
	 * 
	 * @author Karsten Lehmann
	 *
	 */
	private static class ShutdownLogData extends LogData {
		
	}
}
